#!/usr/bin/env python3
import os
import sys
import random
import hashlib
import platform
import pandas as pd

from time import time, strftime, gmtime
from datetime import datetime
from argparse import ArgumentParser, RawTextHelpFormatter, FileType
from functools import partial

from blessed import Terminal


filename = os.path.basename(sys.argv[0])
doc = f"""example:
  {filename} -i test.txt -s -d 60
  echo 'The typing seems really strong today.' | {filename} -d 3.5
  {filename} < test.txt

shortcuts:
  ^c / ctrl+c           end the test and get results now
  ^[ / ctrl+[ / esc     end the test and get results now
  ^h / ctrl+h / bksp    delete a character
  ^r / ctrl+r / tab     restart the same test
  ^s / ctrl+s           restart the test with words reshuffled
  ^w / ctrl+w           delete a word
  ^u / ctrl+u           delete a word
"""


def run():
    main(**parse_args())


def main(duration, input, rows, shuffle, help,
         output, mistyped, char_speeds, word_speeds, hash):
    """Reads test words from `input` delimited by whitespace characters.
    Listens to standard input forming a typed word every time a
    whitespace is pressed. Compares typed words with test words. Updates the
    screen calling `draw` every time a key is pressed or every 0.1 seconds,
    whichever comes first.

    The test ends when `duration` time has passed or all words have been typed.
    Upon exiting, test results are printed and stored in `output`.
    """
    if input.isatty():  # no test words provided, fallback to a default test
        basedir = os.path.dirname(__file__)
        input = open(basedir + '/tests/common_300', 'r')
        duration = 60
        shuffle = True

    test = input.read()
    if hash is None:
        hash = hashlib.sha1(test.encode("utf-8")).hexdigest()
    words = test.split()

    if shuffle:
        random.shuffle(words)

    if not sys.__stdin__.isatty():  # force stdin from user
        if platform.system() == 'Windows':
            sys.__stdin__ = open('con:', 'r')  # NOT TESTED
        else:
            sys.__stdin__ = open('/dev/tty', 'r')

    term = Terminal()

    color_normal = term.normal
    color_correct = term.color_rgb(0, 230, 0)
    color_wrong = term.color_rgb(230, 0, 0)

    correct_chars = total_chars = -1
    wpm = 0
    time_passed = actual_duration = start = 0
    word_i = 0
    text = ''
    colors = [color_normal]*len(words)

    mistyped_words = []
    char_times = []

    with term.raw(), \
            term.cbreak(), \
            term.fullscreen(), \
            term.hidden_cursor():
        while word_i < len(words) and (not start or time() - start < duration):
            word = words[word_i]

            color = color_correct if word == text else \
                color_normal if word.startswith(text) else \
                color_wrong
            colors[word_i] = color + term.reverse

            draw(term, rows, words, colors, word_i, text, wpm, time_passed)

            char = term.inkey(timeout=0.1, esc_delay=0)
            char_time = time()
            time_passed = char_time - start if start else 0

            if not char:
                continue

            if not start:
                start = time()

            if char == '\x03' or char == '\x1b':  # ctrl-c or ctrl-[ or esc
                break

            elif char == '\x08' or char == '\x7f':  # ctrl-h or bksp
                text = text[:-1]

            elif char in ('\x12', '\x13', '\t'):  # ctrl-r or ctrl-s or tab
                correct_chars = total_chars = -1
                wpm = 0
                time_passed = actual_duration = start = 0
                word_i = 0
                text = ''
                colors = [color_normal]*len(words)
                mistyped_words = []
                char_times = []
                if char == '\x13':
                    random.shuffle(words)

            elif char == '\x15' or char == '\x17':  # ctrl-u or ctrl-w
                text = ''

            elif char.isspace() and text:
                if word_i + 1 < len(words):  # if not last space
                    total_chars += 1
                    correct_chars += 1

                if text == word:
                    correct_chars += len(word)
                    colors[word_i] = color_correct
                else:
                    colors[word_i] = color_wrong
                    mistyped_words.append((word, text))

                actual_duration = time_passed
                wpm = min(int(correct_chars*12/actual_duration), 999)

                text = ''
                word_i += 1
                char_times.append((char, char_time))

            elif not char.isspace():
                total_chars += 1
                text += char
                if word_i + 1 >= len(words) and words[-1] == text:
                    term.ungetch(' ')
                char_times.append((char, char_time))

    total_chars -= len(text)

    if actual_duration <= 0 or total_chars <= 0:
        return

    accuracy = 100 * correct_chars // total_chars
    timestamp = datetime.now().strftime("%d/%m/%Y %H:%M:%S")

    print(f'accuracy: {accuracy}%')
    print(f'speed:    {wpm}wpm')
    print(f'duration: {actual_duration:.2f}s')

    chars, times = zip(*char_times)
    durations = [t1-t0 for t0, t1 in zip(times, times[1:])]
    char_durations = list(zip(chars, durations))

    row = [timestamp, wpm, accuracy, actual_duration, duration, hash]
    output.write(','.join(map(str, row)) + '\n')

    df = pd.DataFrame([(w1, w2, timestamp) for w1, w2 in mistyped_words])
    mistyped.write(df.to_csv(index=False, header=False))

    df = pd.DataFrame([(c, t, 12/t, timestamp) for c, t in char_durations])
    char_speeds.write(df.to_csv(index=False, header=False))

    word = ''
    word_i = 0
    word_duration = 0
    word_durations = []
    for char, duration in char_durations:
        if char.isspace():
            if word == words[word_i]:
                word_durations.append((word, word_duration))
            word_i += 1
            word = ''
            word_duration = 0
        else:
            word += char
            word_duration += duration

    df = pd.DataFrame([(word, t, len(word)*12/t, timestamp)
                       for word, t in word_durations])
    word_speeds.write(df.to_csv(header=False, index=False))


def draw(term, rows, words, colors, word_i, text, wpm, time_passed):
    """Text wraps the `words` list to the terminal width, and prints `rows`
    lines of wrapped words coloured with `colors` starting with the line
    containing the current word that is being typed.
    Then, if there is space, prints `prompt` + `text` + `stats`.
    """
    def join(words, length):
        eol = term.clear_eol if length + len(words) - 1 < term.width else ''
        return ' '.join(line_words) + eol

    echo = partial(print, end='', flush=True, file=term.stream)
    allowed_height = min(term.height, rows)

    len_line = 0
    line_words = []
    line_height = None
    for i, (word, color) in enumerate(zip(words, colors)):
        if len_line + len(word) + len(line_words) > term.width:
            if line_height is not None:
                echo(term.move_yx(line_height, 0) + join(line_words, len_line))
                line_height += 1
                if line_height >= allowed_height:
                    break

            len_line = 0
            line_words = []

        if i == word_i:
            line_height = 0

        len_line += len(word)
        line_words.append(color + word + term.normal)

    else:
        echo(term.move_yx(line_height, 0) + join(line_words, len_line))
        line_height += 1

    if allowed_height > 1:
        prompt = '>>>'
        timestamp = strftime('%H:%M:%S', gmtime(time_passed))
        stats = f'{wpm:3d} wpm | {timestamp}'
        n = term.width - len(prompt) - len(stats)
        echo(term.move_yx(line_height, 0) + f'{prompt}{text[:n]: <{n}}{stats}')

    for i in range(1, allowed_height - line_height + 1):
        echo(term.move_yx(line_height + i, 0) + term.clear_eol)


def parse_args():
    """Parses `sys.argv` and returns a dictionary suitable for `main`."""
    parser = ArgumentParser(epilog=doc, formatter_class=RawTextHelpFormatter)

    default = '(default: %(default)s)'
    basedir = os.path.dirname(__file__)
    parser.add_argument('-d', '--duration', type=float, default=float('inf'),
                        help='duration in seconds ' + default)
    parser.add_argument('--hash', type=str, default=None,
                        help='custom hash (generated from input by default)')
    parser.add_argument('-i', '--input', type=FileType('r'),
                        default=sys.stdin,
                        help='file to read words from (default: sys.stdin)')
    parser.add_argument('-o', '--output', type=FileType('a'),
                        default=basedir + '/results/results.csv',
                        help='file to store results in\n' + default)
    parser.add_argument('-m', '--mistyped', type=FileType('a'),
                        default=basedir + '/results/mistyped_words.csv',
                        help='file to store mistyped words in\n' + default)
    parser.add_argument('-c', '--char_speeds', type=FileType('a'),
                        default=basedir + '/results/char_speeds.csv',
                        help='file to store character speeds in\n' + default)
    parser.add_argument('-w', '--word_speeds', type=FileType('a'),
                        default=basedir + '/results/word_speeds.csv',
                        help='file to store word speeds in\n' + default)
    parser.add_argument('-s', '--shuffle', action='store_true',
                        help='shuffle words ' + default)
    parser.add_argument('-r', '--rows', type=int, default=2,
                        help='number of test rows to show ' + default)

    return dict(parser.parse_args()._get_kwargs(), help=parser.print_help)


if __name__ == '__main__':
    main(**parse_args())



#!/usr/bin/env python3
from io import StringIO
from collections import deque, defaultdict
from argparse import ArgumentParser, RawTextHelpFormatter, FileType
from functools import partial
from itertools import cycle
from matplotlib import rcParams

import os
import re
import sys
import math
import warnings
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from typetest.utils import damerau_levenshtein_distance

warnings.simplefilter('ignore', np.RankWarning)
rcParams.update({'figure.autolayout': True})
sns.set(font_scale=1.5)
filename = os.path.basename(sys.argv[0])
doc = f"""example:
  {filename}
  {filename} wpm
  {filename} char word
"""

known_hashes = {
    'da4846a3c2a8469dd77c921ab0b0bcd506b6e9f3':
        '300 most common english words',
    '275eb003c4fba77d7e61893c3d9fa869822e06c8':
        '1000 most common english words (no double letters)'
}


def run():
    main(**parse_args())


def main(graphs, output, mistyped, char_speeds, word_speeds, help):
    is_word = partial(re.match, r'^[a-z]+$')
    if 'wpm' in graphs:
        plot_wpm(output)
    if 'char' in graphs:
        plot_char_speeds(char_speeds, filter_func=str.islower)
    if 'word' in graphs:
        plot_n_best_word_speeds(word_speeds, 50, filter_func=is_word)
    if 'dist' in graphs:
        plot_word_wpm_distribution(word_speeds, filter_func=is_word)
    if 'mistypes' in graphs:
        plot_mistypes_distribution(mistyped, filter_func=is_word)


def plot_wpm(output):
    """Reads `output` and plots typing speeds uniformly apart.
    Adds a trendline.
    """
    df = pd.read_csv(output, header=None,
                     names=['timestamp', 'wpm', 'accuracy',
                            'actual_duration', 'duration', 'hash'])

    df.timestamp = pd.to_datetime(df.timestamp)
    # df = df.set_index(df.timestamp)

    min_wpm = None
    gdf = defaultdict(lambda: [[], pd.DataFrame()])
    for index, row in df.iterrows():
        h = row['hash']
        indexes, hdf = gdf[h]
        indexes.append(index)
        hdf = hdf.append(row)
        gdf[h] = indexes, hdf
        if min_wpm is None or row['wpm'] < min_wpm:
            min_wpm = row['wpm']

    # grouped = sorted(gdf.items(), key=lambda x: x[1][1]['wpm'].mean(),
    #                  reverse=True)
    grouped = gdf.items()

    fig, ax = plt.subplots()
    colors = cycle(sns.color_palette())
    for h, (indexes, hdf) in grouped:
        if h in known_hashes:
            h = known_hashes[h]
        x = indexes
        y = hdf.wpm
        color = next(colors)
        ax.plot(x, y, color=color, lw=3, label=h)
        # ax.fill_between(x, y, min_wpm, facecolor=color, label=h)
        trendline = np.poly1d(np.polyfit(x, y, 1))(x)
        ax.plot(x, trendline, '-', lw=4, color='white')
        ax.plot(x, trendline, '--', lw=2, color=color, label='trendline')

    ax.plot(df.accuracy, color='white', lw=4, alpha=0.5)
    ax.plot(df.accuracy, color=next(colors), lw=1.5, label='accuracy [%]', alpha=0.5)

    ax.set_title('typing speed per typing test')
    ax.set_xlabel('')
    ax.set_ylabel('typing speed [wpm]')
    ax.legend()

    # ticks = plt.yticks()[0]
    # plt.yticks(np.arange(0, ticks[-1], 10))
    plt.xticks(df.index, df.timestamp.dt.date, rotation=90)
    # label only every 50th tick
    for i, label in enumerate(ax.xaxis.get_ticklabels()):
        if i % math.ceil(len(df) / 50):
            label.set_visible(False)

    plt.show()


def plot_char_speeds(char_speeds, size=10000, filter_func=lambda c: True):
    """Reads last `size` lines of `char_speeds` and groups them by characters.
    Removes lowest and highest 10% and boxplots the data.

    filter_func: function taking a `char` returning `True` if char should be
    plotted, `False` otherwise. By default plots all characters.
    """
    q = deque(char_speeds, maxlen=size)
    df = pd.read_csv(StringIO(''.join(q)), header=None,
                     names=['char', 'duration', 'wpm', 'timestamp'])

    gdf = filter(lambda t: filter_func(t[1]['char'].iloc[0]),
                 df.groupby(['char']))
    wpms = []
    chars = []
    means = []
    for char, df in gdf:
        if filter_func(char):
            q1 = df['wpm'].quantile(0.1)  # noqa
            q3 = df['wpm'].quantile(0.9)  # noqa
            wpm = df.query('@q1 <= wpm <= @q3')['wpm']
            chars.append(char)
            wpms.append(wpm)
            means.append(wpm.mean())

    assert chars, 'Not enough data'
    fig, ax = plt.subplots()

    ax.boxplot(wpms, labels=chars)
    mean = round(sum(means)/len(means))
    ax.axhline(y=mean, color='r', linestyle='-', label=f'mean {mean} wpm')

    ax.set_title(f'typing speed per character of last {size} characters')
    ax.set_xlabel('')
    ax.set_ylabel('typing speed [wpm]')
    ax.legend()

    ticks = plt.yticks()[0]
    plt.yticks(np.arange(0, ticks[-1], 10))

    plt.show()


def plot_n_best_word_speeds(word_speeds, n, filter_func=lambda w: True):
    """Loads all words from `word_speeds` and groups them by word.
    """
    half = n//2
    df = pd.read_csv(word_speeds, header=None,
                     names=['word', 'duration', 'wpm', 'timestamp'])

    gdf = list(filter(lambda t: filter_func(t[0]), df.groupby(['word'])))

    with open('least_typed_words', 'w') as f:
        words_by_freq = list(zip(*sorted(gdf, key=lambda t: len(t[1]))))[0]
        f.write(' '.join(words_by_freq))

    gdf = sorted(gdf, key=lambda x: x[1]['wpm'].median())

    with open('worst_words', 'w') as f:
        words_by_median = list(zip(*gdf))[0]
        f.write(' '.join(words_by_median))

    first_half = deque(maxlen=half)
    second_half = deque(maxlen=half)
    for word, df in gdf:
        if len(first_half) < first_half.maxlen:
            first_half.append((word, df['wpm'], df['wpm'].mean()))
        else:
            second_half.append((word, df['wpm'], df['wpm'].mean()))

    words, wpms, means = zip(*list(first_half) + list(second_half))
    mean = round(sum(means)/len(means))

    assert words, 'Not enough data'

    fig, ax = plt.subplots()

    ax.boxplot(wpms, labels=words)
    ax.axhline(y=mean, color='r', linestyle='-', label=f'mean {mean} wpm')

    ax.set_title(f'worst and best {half} words')
    ax.set_xlabel('')
    ax.set_yscale('linear')
    ax.set_ylabel('typing speed [wpm]')
    ax.legend()

    ticks = plt.yticks()[0]
    plt.yticks(np.arange(ticks[0], ticks[-1], 10))
    plt.xticks(rotation=90)

    plt.show()


def plot_word_wpm_distribution(word_speeds, filter_func=lambda c: True):
    df = pd.read_csv(word_speeds, header=None,
                     names=['word', 'duration', 'wpm', 'timestamp'])

    gdf = list(filter(lambda t: filter_func(t[0]), df.groupby(['word'])))
    wpms = [df['wpm'].median() for word, df in gdf]

    ax = sns.histplot(wpms, kde=True, stat='probability')
    ax.set_title('percentage of words typed at a certain speed')
    ax.set_xlabel('typing speed in wpm')
    ax.set_ylabel('percentage of words')
    plt.show()


def plot_mistypes_distribution(mistyped, filter_func=lambda c: True):
    df = pd.read_csv(mistyped, header=None,
                     names=['word', 'mistype', 'timestamp'])

    mistakes = defaultdict(int)
    words, mistypes = df['word'], df['mistype']
    for word, mistype in zip(words, mistypes):
        if abs(len(word) - len(mistype)) < 10:
            distance = damerau_levenshtein_distance(word, mistype)
            if distance >= max(len(word), len(mistype)):
                # wrong word has been typed
                continue
            if distance > 3 and word.startswith(mistype):
                # word might be formed of multiple words
                # if mistype in dictionary:
                continue

            mistakes[distance] += 1

    fig, ax = plt.subplots()
    labels, sizes = zip(*sorted(mistakes.items()))
    explode = [0] + [0.2]*(len(mistakes) - 1)
    ax.pie(sizes, labels=labels, autopct='%1.1f%%',
           explode=explode)
    # ax = sns.histplot(mistakes, stat='probability')
    ax.set_title('number of mistakes made when typing a word')
    plt.show()


def parse_args():
    """Parses `sys.argv` and returns a dictionary suitable for `main`."""
    parser = ArgumentParser(epilog=doc, formatter_class=RawTextHelpFormatter)

    default = '(default: %(default)s)'
    basedir = os.path.dirname(__file__)
    resultsdir = 'results'
    parser.add_argument('graphs', type=str, nargs='*',
                        default=['wpm'],
                        help='graphs to plot: wpm char word dist mistypes\n' + default)
    parser.add_argument('-o', '--output', type=FileType('r'),
                        default=f'{basedir}/{resultsdir}/results.csv',
                        help='file to store results in\n' + default)
    parser.add_argument('-m', '--mistyped', type=FileType('r'),
                        default=f'{basedir}/{resultsdir}/mistyped_words.csv',
                        help='file to store mistyped words in\n' + default)
    parser.add_argument('-c', '--char_speeds', type=FileType('r'),
                        default=f'{basedir}/{resultsdir}/char_speeds.csv',
                        help='file to store character speeds in\n' + default)
    parser.add_argument('-w', '--word_speeds', type=FileType('r'),
                        default=f'{basedir}/{resultsdir}/word_speeds.csv',
                        help='file to store word speeds in\n' + default)

    return dict(parser.parse_args()._get_kwargs(), help=parser.print_help)


if __name__ == '__main__':
    main(**parse_args())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['typetest']

package_data = \
{'': ['*'], 'typetest': ['results/placeholder', 'tests/*']}

install_requires = \
['blessed>=1.18.1,<2.0.0',
 'matplotlib>=3.4.2,<4.0.0',
 'pandas>=1.3.0,<2.0.0',
 'seaborn>=0.11.1,<0.12.0']

entry_points = \
{'console_scripts': ['typetest = typetest.__main__:run',
                     'typetest-analyse = typetest.analyse:run']}

setup_kwargs = {
    'name': 'typetest',
    'version': '0.1.4',
    'description': 'Test your typing speed without leaving the terminal.',
    'long_description': '<p align="center">\n  <img src="https://user-images.githubusercontent.com/16375100/125825025-74b5b1cd-c5d2-40f1-902a-5b5902720d90.png" width="300"/>\n</p>\n<p align="center">Test your typing speed without leaving the terminal.</p>\n\n<p align="center">\n  <a href="https://github.com/mastermedo/typetest">\n    <img src="https://img.shields.io/github/languages/code-size/mastermedo/typetest" alt="build" title="build"/>\n  </a>\n  <a href="https://github.com/mastermedo/typetest/LICENSE">\n    <img src="https://img.shields.io/github/license/mastermedo/typetest" alt="license" title="license"/>\n  </a>\n  <a href="https://github.com/mastermedo/typetest/stargazers">\n    <img src="https://img.shields.io/badge/maintainer-mastermedo-yellow" alt="maintainer" title="maintainer"/>\n  </a>\n</p>\n\n<p align="center">\n  <a href="https://github.com/mastermedo/typetest">\n    <img src="./img/typetest-demo.gif" alt="demo" title="demo"/>\n  </a>\n</p>\n\n## :clipboard: description\n`typetest` is a self-contained minimal typing test program written with [blessed](https://github.com/jquast/blessed/).\nAs is, it is a near clone of [10fastfingers](https://10fastfingers.com/typing-test/english) with an added bonus of being able to see typing speed as you\'re typing.\n\n## :zap: features\n1. adjustable settings\n2. storing test results\n3. analysing mistakes\n4. easy to track improvement\n\n## :chart_with_upwards_trend: analyse test results!\n![wpm](https://user-images.githubusercontent.com/16375100/125824726-6304ee64-ddf1-4456-879c-10daca45d91c.png)\n[char_speeds](https://user-images.githubusercontent.com/16375100/125824817-5c2cbcae-fdcc-45c9-9a3b-ed5c3ec497a5.png)\n![word_speeds](https://user-images.githubusercontent.com/16375100/125824889-a01bb4bb-1ed2-49ed-a0aa-9bd5f6b411c7.png)\n![mistypes](https://user-images.githubusercontent.com/16375100/125824921-3ecdf9f4-804e-41ec-98a4-6343d0ffbbe2.png)\n![dist](https://user-images.githubusercontent.com/16375100/125824933-01294d91-92c9-4ae0-9910-539f6d16507e.png)\n\n## :shipit: installation\n\n1. install python 3\n2. install [blessed](https://pypi.org/project/blessed/)\n3. clone this repository\n4. run `python typetest -s -d 60 < common_300`\n5. (optional) add `typetest` to path or make an alias like `tt`\n6. (optional) store your results in some file and analyse\n\n## :bulb: ideas for tests\nAlong with `typetest` this repository features sample tests.\nTry them like so: `typetest -s -d 60 -i common_200` or scrape something off the internet, like a [featured article](https://en.wikipedia.org/wiki/Wikipedia:Featured_articles) on wikipedia.\n\n```python\n#!/usr/bin/env python3\nimport re\nimport requests\nfrom bs4 import BeautifulSoup\n\nword_pattern = re.compile(r"[\'A-Za-z\\d\\-]+[,\\.\\?\\!]?")  # symbols to keep\nurl = \'https://en.wikipedia.org/wiki/Special:RandomInCategory/Featured_articles\'\n\nr = requests.get(url)\nsoup = BeautifulSoup(r.text, \'html.parser\')\nfor sup in soup.select(\'sup\'):\n    sup.extract()  # remove citations\n\ntext = \' \'.join(p.text for p in soup.select(\'p\'))\ntext = re.sub(r\'\\[.*?\\]|\\(.*?\\)\', \'\', text)  # remove parenthesis\nprint(\' \'.join(re.findall(word_pattern, text)))\n```\nIf you create a file called `wiki_random` you can start the test with `wiki_random | typetest`.\nWrite your own scraper, you may find some suggestions [here](https://en.wikipedia.org/wiki/Lists_of_English_words).\n\n## :question: usage\n\n```\nusage: typetest [-h] [-d DURATION] [-i INPUT] [-o OUTPUT] [-s] [-r ROWS]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -d DURATION, --duration DURATION\n                        duration in seconds (default: inf)\n  -i INPUT, --input INPUT\n                        file to read words from (default: sys.stdin)\n  -o OUTPUT, --output OUTPUT\n                        file to store results in\n                        (default: /home/medo/repos/typetest/results)\n  -s, --shuffle         shuffle words (default: False)\n  -r ROWS, --rows ROWS  number of test rows to show (default: 2)\n\nexample:\n  typetest -i test.txt -s -d 60\n  echo \'The typing seems really strong today.\' | typetest -d 3.5\n  typetest < test.txt\n\nshortcuts:\n  ^c / ctrl+c           end the test and get results now\n  ^h / ctrl+h           backspace\n  ^r / ctrl+r           restart the same test\n  ^w / ctrl+w           delete a word\n  ^u / ctrl+u           delete a word\n```\n\n<p align="center">\n  <a href="#">\n    <img src="https://img.shields.io/badge/⬆️back_to_top_⬆️-white" alt="Back to top" title="Back to top"/>\n  </a>\n</p>\n',
    'author': 'MasterMedo',
    'author_email': 'mislav.vuletic@gmail.com',
    'maintainer': 'MasterMedo',
    'maintainer_email': 'mislav.vuletic@gmail.com',
    'url': 'https://github.com/MasterMedo/typetest',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



def damerau_levenshtein_distance(word1: str, word2: str) -> int:
    inf = len(word1) + len(word2)
    table = [[inf for _ in range(len(word1) + 2)]
             for _ in range(len(word2) + 2)]

    for i in range(1, len(word1) + 2):
        table[1][i] = i - 1
    for i in range(1, len(word2) + 2):
        table[i][1] = i - 1

    da = {}
    for col, c1 in enumerate(word1, 2):
        last_row = 0
        for row, c2 in enumerate(word2, 2):
            last_col = da.get(c2, 0)

            addition = table[row-1][col] + 1
            deletion = table[row][col-1] + 1
            substitution = table[row-1][col-1] + (0 if c1 == c2 else 1)

            transposition = table[last_row-1][last_col-1]\
                + (col - last_col - 1)\
                + (row - last_row - 1)\
                + 1

            table[row][col] = min(
                addition,
                deletion,
                substitution,
                transposition
            )

            if c1 == c2:
                last_row = row
        da[c1] = col

    return table[len(word2)+1][len(word1)+1]
